/**
 * \file main_pmsm.c
 * \author Martin Prudek
 * \brief Mainfile pro pmsm control.
 */

#ifndef NULL
#define NULL (void*) 0
#endif /*NULL*/


#include <stdlib.h> 	/*exit*/
#include <signal.h> 	/*signal handler Ctrl+C*/
#include <stdio.h> 	/*printf*/
#include <sched.h>	/*sheduler*/
#include <unistd.h>	/*usleep*/
#include <pthread.h>	/*threads*/
#include <time.h>	/*nanosleep*/
#include <string.h>

#include "rpin.h"	/*gpclk*/
#include "rp_spi.h"	/*spi*/
#include "misc.h" 	/*structure for priorities*/
#include "pmsm_state.h"
#include "cmd_proc.h"
#include "controllers.h"
#include "commutators.h"




#define PRIOR_KERN	50
#define PRIOR_HIGH	49
#define PRIOR_LOW	20

#define THREAD_SHARED 	0
#define INIT_VALUE	1 	/*init value for semaphor*/


#define NSEC_PER_SEC    (1000000000) /* The number of nsecs per sec. */


struct rpi_in data;
struct rpi_state rps={
	//.MAX_DUTY=170,
	.spi_dat=&data,
	.test=0,
	.pwm1=0,.pwm2=0, .pwm3=0,
	.pwm1=0, .t_pwm2=0, .t_pwm3=0,
	.commutate=0,
	.duty=0,			/* duty cycle of pwm */
	.index_dist=0,		/* distance to index position */
	.index_ok=0,
	.tf_count=0,		/*number of transfer*/
	.desired_pos=0,		/* desired position */
	.pos_reg_ena=0,
	.desired_spd=0,
	.spd_reg_ena=0,
	.old_pos={0},
	.spd_err_sum=0,
	.log_col_count=0,	/* pocet radku zaznamu */
	.log_col=0,
	.doLogs=0,
	.alpha_offset=960
};

/**
 * \brief Initilizes GPCLK.
 */
int clk_init()
{
	initialise(); /*namapovani gpio*/
	initClock(PLLD_500_MHZ, 10, 0);
	gpioSetMode(4, FSEL_ALT0);
	return 0;
}
/*
 * \brief Terminates GPCLK.
 */

inline void clk_disable(){
	termClock(0);
}


/*
 * \brief
 * Free logs
 */
void freeLogs(){
	int r;
	if (rps.log_col_count){
		for (r=0;r<LOG_ROWS;r++){
			free(rps.logs[r]);
		}
	}
	rps.log_col_count=0;
	rps.doLogs=0;
}

/*
 * \brief
 * Makes log.
 */
void makeLog(){
	int r;
	if (rps.log_col==MAX_LOGS-1){
		rps.doLogs=0;
		return;
	}
	rps.logs[0][rps.log_col]=(int)rps.tf_count;
	rps.logs[1][rps.log_col]=(int)rps.spi_dat->pozice;

	rps.logs[2][rps.log_col]=(int)rps.pwm1;
	rps.logs[3][rps.log_col]=(int)rps.pwm2;
	rps.logs[4][rps.log_col]=(int)rps.pwm3;
	rps.logs[5][rps.log_col]=rps.duty;

	rps.logs[6][rps.log_col]=rps.desired_spd;
	rps.logs[7][rps.log_col]=rps.speed;

	rps.logs[8][rps.log_col]=(int)(rps.spi_dat->ch1/rps.spi_dat->adc_m_count);
	rps.logs[9][rps.log_col]=(int)(rps.spi_dat->ch2/rps.spi_dat->adc_m_count);
	rps.logs[10][rps.log_col]=(int)(rps.spi_dat->ch0/rps.spi_dat->adc_m_count);

	rps.log_col++;
     /*
        if (rps.log_col==rps.log_col_count-1){
        	rps.log_col_count*=2;
        	rps.log_col_count%=MAX_LOGS;
        	for (r=0;r<LOG_ROWS;r++){
        		rps.logs[r]=realloc(rps.logs[r],rps.log_col_count*sizeof(int));
        		if (rps.logs[r]==NULL){
        			rps.doLogs=0;
        			rps.error=1;
        		}
        	}
        }
     */
}




/**
 * \brief Signal handler pro Ctrl+C
 */
void appl_stop(){
	uint8_t tx[16];
	sem_wait(&rps.thd_par_sem);

	memset(tx,0,16*sizeof(int));
	rps.pwm1=0;
	rps.pwm2=0;
	rps.pwm3=0;
	spi_read(&rps);

	spi_disable();
	clk_disable();
	freeLogs();
	/*muzeme zavrit semafor*/
	sem_destroy(&rps.thd_par_sem);
        printf("\nprogram bezpecne ukoncen\n");
}

void substractOffset(struct rpi_in* data, struct rpi_in* offset){
	data->pozice=data->pozice_raw-offset->pozice_raw;
	return;
}



/**
 * Funkce pravidelne vypisuje posledni zjistenou pozici lokalniho motoru
 */
void * pos_monitor(void* param){
	while(1){
		printData(&rps);
		usleep(1000000); 	/*1 Hz*/
	}
	return (void*)0;
}
/*
 * \brief
 * Multiplication of 11 bit
 * Zaporne vysledky prvede na nulu.
 */
inline uint16_t mult_cap(int32_t s,int d){
	int j;
	int res=0;
	for(j=0;j!=11;j++){
		/* multiplicate as if maximum sinus value was unity */
		res+=(!(s & 0x10000000))*(((1 << j) & s)>>j)*(d>>(10-j));
	}
	return res;
}


/**
 * \brief
 * Computation of distance to index.
 *
 * K dispozici je 12-bit index, to umoznuje ulozit 4096 ruznych bodu
 * Je nutne vyjadrit 1999 bodu proti i posmeru h.r. od indexu -
 * 	to je 3999 bodu
 * 	=>12 bitu je dostacujicich, pokud nikdy nedojde ke ztrate
 * 		signalu indexu
 */
void comIndDist(){
	uint16_t pos = 0x0FFF & data.pozice_raw;
	uint16_t dist;
	uint16_t index = data.index_position;

	if (index<1999){		/*index e<0,1998> */
		if (pos<index){			/*pozice e<0,index-1> */
			/*proti smeru h.r. od indexu*/
			dist=pos+2000-index;
		}else if (pos<=index+1999){	/*pozice e<index,index+1999> */
			/*po smeru h.r. od indexu*/
			dist=pos-index;
		}else if (pos<index+2096){	/*pozice e<index+2000,index+2095> */
			goto index_lost;
		}else{				/*pozice e<index+2096,4095> */
			/*proti smeru h.r. od indexu - podtecena pozice*/
			dist=pos-index-2096;
		}
	}else if (index<=2096){		/*index e<1999,2096>*/
		if (pos<index-1999){		/*pozice e<0,index-2000> */
			goto index_lost;
		}else if (pos<index){		/*pozice e<index-1999,index-1> */
			/*proti smeru h.r. od indexu*/
			dist=pos+2000-index;
		}else if (pos<=index+1999){	/*pozice e<index,index+1999> */
			/*po smeru h.r. od indexu*/
			dist=pos-index;
		}else {				/*pozice e<index+2000,4095> */
			goto index_lost;
		}
	}else{				/*index e<2097,4095> */
		if (pos<=index-2097){		/*pozice e<0,index-2097> */
			/*po smeru h.r. od indexu - pretecena pozice*/
			dist=4096+pos-index;
		}else if (pos<index-1999){	/*pozice e<index-2096,index-2000> */
			goto index_lost;
		}else if (pos<index){		/*pozice e<index-1999,index-1> */
			/*proti smeru h.r. od indexu*/
			dist=pos+2000-index;
		}else{ 				/*pozice e<index,4095> */
			/*po smeru h.r. od indexu*/
			dist=pos-index;
		}
	}

	rps.index_dist = dist;
	return;

	index_lost:
		rps.index_ok=0;
		return;
}


/*
 * \brief
 * Computate speed.
 */
void compSpeed(){
	signed long int spd;
	spd=rps.spi_dat->pozice-rps.old_pos[rps.tf_count%OLD_POS_NUM];
	rps.speed=(int32_t)spd;
}



/*
 * \brief
 * Feedback loop.
 * TODO: replace bunch of 'IFs' with Object-like pattern
 */
void * read_data(void* param){
	int i;
	struct rpi_in pocatek;
	struct rpi_state poc={
		.spi_dat=&pocatek,
		.test=0,
		.pwm1=0, .pwm1=0, .pwm3=0
	};
	struct timespec t;
	int interval = 1000000; /* 1ms ~ 1kHz*/
	char first=1;
	uint16_t last_index;				/*we have index up-to date*/
	spi_read(&poc);		/*pocatecni informace*/
	clock_gettime(CLOCK_MONOTONIC ,&t);
	/* start after one second */
	t.tv_sec++;
		while(1){
			/* wait until next shot */
			clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &t, NULL);
			sem_wait(&rps.thd_par_sem);		/*---take semaphore---*/

			/*old positions*/
			rps.old_pos[rps.tf_count%OLD_POS_NUM]=rps.spi_dat->pozice;
			spi_read(&rps);		/*exchange data*/
			/*subtract initiate postion */
			rps.tf_count++;
			substractOffset(&data,poc.spi_dat);
			compSpeed();			/*spocita rychlost*/

			if (!rps.index_ok){
				if (first){
					last_index=data.index_position;
					first=0;
				}else if (last_index!=data.index_position){
					rps.index_ok=1;
					comIndDist();	/*vypocet vzdalenosti indexu*/
				}
			}else{ /*index je v poradku*/
				comIndDist();		/*vypocet vzdalenosti indexu*/
			}

			/* pocitame sirku plneni podle potreb rizeni*/
			if (rps.pos_reg_ena){           /*pozicni rizeni*/
				pos_pid(&rps);
			}else if(rps.spd_reg_ena){	/*rizeni na rychlost*/
				spd_pid(&rps);
			}

			/* sirka plneni prepoctena na jednotlive pwm */
			if (rps.index_ok && rps.commutate){
				/*simple_ind_dist_commutator(rps.duty);*/
				/*sin_commutator(rps.duty);*/
				inv_trans_comm(&rps);
				inv_trans_comm_2(&rps);
			}else if(!rps.index_ok && rps.commutate){
				simple_hall_commutator(&rps);
			}

			/*zalogujeme hodnoty*/
			if (rps.doLogs && !(rps.tf_count%LOG_PERIOD)){
				makeLog();
			}

			sem_post(&rps.thd_par_sem);		/*--post semaphore---*/

			/* calculate next shot */
			t.tv_nsec += interval;

			while (t.tv_nsec >= NSEC_PER_SEC) {
				t.tv_nsec -= NSEC_PER_SEC;
				t.tv_sec++;
			}

		}
}


/**
 * \brief Main function.
 */

int main(){
	pthread_t base_thread_id;
	clk_init(); 		/* inicializace gpio hodin */
	spi_init(); 		/* iniicializace spi*/

	/*semafor pro detekci zpracovani parametru vlaken*/
	sem_init(&rps.thd_par_sem,THREAD_SHARED,INIT_VALUE);
	setup_environment();

	base_thread_id=pthread_self();

	/*main control loop*/
	create_rt_task(&base_thread_id,PRIOR_HIGH,read_data,NULL);

	/*monitor of current state*/
	create_rt_task(&base_thread_id,PRIOR_LOW,pos_monitor,NULL);

	/*wait for commands*/
	poll_cmd(&rps);

	return 0;
}
